﻿using System;
using System.IO;
using System.Linq;
using System.Net.Security;
using System.ServiceModel;
using System.Xml;
using System.Xml.Serialization;
using ArgusTV.DataContracts;
using ArgusTV.ServiceAgents;
using ArgusTV.ServiceContracts;
using JMS.DVB;
using JMS.DVBVCR.RecordingService.ProgramGuide;


namespace SampleArgusTVRecordingService
{
    /// <summary>
    /// Repräsentiert den Windows Prozess als Ganzes.
    /// </summary>
    public static class Program
    {
        /// <summary>
        /// Der Name des zu verwendenden Geräteprofils.
        /// </summary>
        private static string DVBNETProfile { get { return Properties.Settings.Default.Profiles.Split( ':' ).First(); } }

        /// <summary>
        /// Globale initialisierung.
        /// </summary>
        static Program()
        {
            // For guide import we use DVB.NET
            RunTimeLoader.Startup();
        }

        /// <summary>
        /// Startet die Anwendung.
        /// </summary>
        /// <param name="args">Feinsteurung der Anwendung.</param>
        public static void Main( string[] args )
        {
            // Dispatch
            if (args.Length == 0)
                RunService();
            else if (args.Length == 1)
                switch (args[0])
                {
                    case "/prepare": ImportProgramGuide(); break;
                }
        }

        /// <summary>
        /// Führt den Dienst aus.
        /// </summary>
        private static void RunService()
        {
            // Start host
            using (var host = new ServiceHost( new RecordingService() ))
            {
                // Create binding
                var binding =
                    new NetTcpBinding( SecurityMode.None )
                    {
                        MaxReceivedMessageSize = 256 * 1024 * 1024,
                        ReceiveTimeout = new TimeSpan( 0, 30, 0 ),
                        SendTimeout = new TimeSpan( 0, 30, 0 ),
                        Namespace = "http://www.argus-tv.com",
                        ReaderQuotas =
                            new XmlDictionaryReaderQuotas
                            {
                                MaxStringContentLength = int.MaxValue,
                                MaxDepth = int.MaxValue,
                                MaxArrayLength = int.MaxValue,
                                MaxNameTableCharCount = int.MaxValue,
                                MaxBytesPerRead = int.MaxValue,
                            },
                    };

                // Configure binding
                binding.Security.Transport.ClientCredentialType = TcpClientCredentialType.None;
                binding.Security.Message.ClientCredentialType = MessageCredentialType.None;
                binding.Security.Transport.ProtectionLevel = ProtectionLevel.None;

                // Start 
                host.AddServiceEndpoint( typeof( IRecorderTunerService ), binding, string.Format( "net.tcp://localhost:{0}/RecorderTunerService", Properties.Settings.Default.ServicePort ) );
                host.Open();

                // Well for now we work manual
                Console.WriteLine( "Running" );
                Console.ReadLine();
            }
        }

        /// <summary>
        /// Erstellt neue Einträge für die Programmzeitschift.
        /// </summary>
        private static void ImportProgramGuide()
        {
            // Attach to DVB.NET profile
            var profile = ProfileManager.FindProfile( DVBNETProfile );
            if (profile == null)
                return;

            // Attach to the path of VCR.NET
            var programFiles = Environment.GetEnvironmentVariable( "ProgramFiles(x86)" ) ?? Environment.GetEnvironmentVariable( "ProgramFiles" );
            var guideFile = Path.Combine( programFiles, string.Format( @"JMS\VCR.NET\Jobs\EPG\EPGData for {0}.xml", DVBNETProfile ) );
            if (!File.Exists( guideFile ))
                return;

            // The manager
            ProgramGuideEntries guide;

            // Attach to stream
            using (var guideStream = new FileStream( guideFile, FileMode.Open, FileAccess.Read, FileShare.Read ))
            using (var guideReader = XmlReader.Create( guideStream, new XmlReaderSettings { CheckCharacters = false } ))
            {
                // Create serializer 
                var serializer = new XmlSerializer( typeof( ProgramGuideEntries ) );

                // Load
                guide = (ProgramGuideEntries) serializer.Deserialize( guideReader );
            }

            // Connect to service
            using (var guideService = new GuideServiceAgent())
            using (var schedulerService = new SchedulerServiceAgent())
                foreach (var entries in guide.Events.GroupBy( ev => ev.Source ))
                {
                    // Locate it
                    var sources = profile.FindSource( entries.Key );
                    if (sources.Length != 1)
                        continue;

                    // Attach to detail information and process
                    var station = (Station) sources[0].Source;
                    var guideChannel = guideService.EnsureChannel( SourceIdentifier.ToString( station ), station.Name, ChannelType.Television );

                    // Check for channels
                    var channels = schedulerService.GetChannelsForGuideChannel( guideChannel );
                    if (channels.Length < 1)
                    {
                        // Create a new channel
                        var channel =
                            new Channel
                            {
                                ChannelType = ChannelType.Television,
                                GuideChannelId = guideChannel,
                                DisplayName = station.Name,
                                Version = 1,
                            };

                        // Remember
                        channel = schedulerService.SaveChannel( channel );

                        // Connect
                        schedulerService.AttachChannelToGuide( channel.ChannelId, guideChannel );
                    }

                    // Extract the related entries
                    var programs = entries.Select( entry => CreateProgram( entry, guideChannel ) );

                    // Report
                    Console.WriteLine( "{1:HH:mm:ss} Importing Guide Programs for {0}", station.FullName, DateTime.Now );

                    // Better add per source since ArgusTV is failry slow on importin guide programs
                    guideService.ImportPrograms( programs.ToArray(), GuideSource.DvbEpg );
                }
        }

        /// <summary>
        /// Erstellt aus einem VCR.NET Eintrag einen ArgusTV Eintrag.
        /// </summary>
        /// <param name="entry">Der VCR.NET Eintrag zu einer Sendung.</param>
        /// <param name="guideChannel">Der zugehörige Sender.</param>
        /// <returns>Der äquivalente Eintrag für ArgusTV.</returns>
        private static GuideProgram CreateProgram( ProgramGuideEntry entry, Guid guideChannel )
        {
            // Process
            return new GuideProgram
            {
                StartTime = entry.StartTime.ToLocalTime(),
                StopTime = entry.EndTime.ToLocalTime(),
                Description = entry.Description,
                StartTimeUtc = entry.StartTime,
                GuideChannelId = guideChannel,
                StopTimeUtc = entry.EndTime,
                Title = entry.Name,
            };
        }
    }
}
