﻿using System;
using System.Collections.Concurrent;
using System.Diagnostics;
using System.Linq;


namespace SampleArgusTVRecordingService
{
    /// <summary>
    /// Alle gerade aktiven Aufzeichnungen.
    /// </summary>
    public class RecordingActivities
    {
        /// <summary>
        /// Alle bekannten Aufzeichnungen.
        /// </summary>
        private readonly ConcurrentDictionary<Guid, RecordingActivity> m_activities = new ConcurrentDictionary<Guid, RecordingActivity>();

        /// <summary>
        /// Ermittelt eine Aktivität und legt sie bei Bedarf neu an.
        /// </summary>
        /// <param name="activityIdentifier">Die eindeutige Kennung der Aktivität.</param>
        /// <param name="service">Der zugehörige Dienst.</param>
        /// <returns>Die gewünschte Aktivität.</returns>
        public RecordingActivity GetOrCreate( Guid activityIdentifier, RecordingService service )
        {
            // Process
            return m_activities.GetOrAdd( activityIdentifier, identifier => new RecordingActivity( service ) );
        }

        /// <summary>
        /// Ermittelt eine Aktivität.
        /// </summary>
        /// <param name="activityIdentifier">Die eindeutige Kennung der Aktivität.</param>
        /// <param name="service">Der zugehörige Dienst.</param>
        /// <returns>Die gewünschte Aktivität.</returns>
        public RecordingActivity Get( Guid activityIdentifier )
        {
            // Process
            RecordingActivity activity;
            if (m_activities.TryGetValue( activityIdentifier, out activity ))
                return activity;
            else
                return null;
        }

        /// <summary>
        /// Entfernt eine Arbeitseinheit.
        /// </summary>
        /// <param name="key">Die eindeutige Kennung der Arbeitseinheit.</param>
        private void Remove( Guid key )
        {
            // Force remove
            RecordingActivity activity;
            m_activities.TryRemove( key, out activity );
        }

        /// <summary>
        /// Wird periodisch aufgerufen und prüft, ob etwas zu tun ist.
        /// </summary>
        public void Tick()
        {
            // Processing limit
            var now = DateTime.UtcNow;

            // Inspect 
            foreach (var activity in m_activities.Where( activity => activity.Value.NextTime <= now ).ToArray())
                try
                {
                    // Process
                    lock (activity.Value.Synchronizer)
                        if (activity.Value.Run())
                            continue;

                    // Discard
                    Remove( activity.Key );
                }
                catch (Exception e)
                {
                    // Force remove
                    Remove( activity.Key );

                    // Just report
                    Trace.TraceError( e.Message );
                }
        }
    }
}
