﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.ServiceModel;
using System.Threading;
using ArgusTV.DataContracts;
using ArgusTV.DataContracts.Tuning;
using ArgusTV.ServiceAgents;
using ArgusTV.ServiceContracts;
using JMS.DVB;


namespace SampleArgusTVRecordingService
{
    /// <summary>
    /// Ein Beispieldienst für Aufzeichnungen mit <i>ArgusTV</i>.
    /// </summary>
    [ServiceBehavior( ConcurrencyMode = ConcurrencyMode.Reentrant, InstanceContextMode = InstanceContextMode.Single )]
    public class RecordingService : IRecorderTunerService, IDisposable
    {
        /// <summary>
        /// Alle unterstützten Geräte.
        /// </summary>
        public readonly RecordingDevices Devices = new RecordingDevices();

        /// <summary>
        /// Der (eindeutige) Name dieses Dienstes.
        /// </summary>
        public string Name { get { return "JMS"; } }

        /// <summary>
        /// Die eindeutige Kennung dieses Dienstes.
        /// </summary>
        public Guid Identifier { get; private set; }

        /// <summary>
        /// Prüft periodisch, ob weitere Aufgaben vorliegen.
        /// </summary>
        private volatile Timer m_timer;

        /// <summary>
        /// Alle bekannten Aktivitäten.
        /// </summary>
        private readonly RecordingActivities m_activities = new RecordingActivities();

        /// <summary>
        /// Erstellt einen neuen Dienst.
        /// </summary>
        public RecordingService()
        {
            // Install timer
            m_timer = new Timer( Tick, null, 1000, Timeout.Infinite );
        }

        /// <summary>
        /// Wird periodisch aufgerufen.
        /// </summary>
        /// <param name="state">Wird ignoriert.</param>
        private void Tick( object state )
        {
            // Make sure that we reactivate
            try
            {
                // Forward
                m_activities.Tick();
            }
            catch (Exception e)
            {
                // Report silently
                Trace.TraceError( e.Message );
            }
            finally
            {
                // See if we are still active
                var timer = m_timer;
                if (timer != null)
                    try
                    {
                        // Reset timer
                        timer.Change( 1000, Timeout.Infinite );
                    }
                    catch (Exception)
                    {
                        // May fail during shutdown
                    }
            }
        }

        /// <summary>
        /// Beendet eine Aufzeichnung.
        /// </summary>
        /// <param name="serverHostName">Der Name des Rechners mit dem ArgusTV Scheduler.</param>
        /// <param name="tcpPort">Die Adresse des Schedulers.</param>
        /// <param name="recordingProgram">Das zu beendende Programm.</param>
        /// <returns>Gesetzt, wenn der Aufruf erfolgreich war.</returns>
        public bool AbortRecording( string serverHostName, int tcpPort, UpcomingProgram recordingProgram )
        {
            // Locate
            var activity = m_activities.Get( recordingProgram.UpcomingProgramId );
            if (activity == null)
                return false;

            // Forward
            return activity.Abort();
        }

        /// <summary>
        /// Prüft, ob eine Ausführung möglich ist.
        /// </summary>
        /// <param name="channel">Der Sender, von dem aufgezeichnet werden soll.</param>
        /// <param name="alreadyAllocated">Alle bereits geplanten Aufzeichnungen.</param>
        /// <param name="useReversePriority">Gesetzt, um die Planungspriorität zu verändern (wird hier ignoriert).</param>
        /// <returns>Der eindeutige Name des Gerätes, auf dem eine Aufzeichnung stattfinden könnte.</returns>
        public string AllocateCard( Channel channel, CardChannelAllocation[] alreadyAllocated, bool useReversePriority )
        {
            // Find all devices which are able to handle the channel of interest
            var candidates =
                new HashSet<string>(
                    Devices
                        .Where( d => d.Profile.FindSource( channel.DisplayName, SourceNameMatchingModes.Name ).Length > 0 )
                        .Select( d => d.Name ), Devices.NameComparer );

            // Check all allocations
            foreach (var allocation in alreadyAllocated)
            {
                // No longer a candidate
                if (!candidates.Contains( allocation.CardId ))
                    continue;

                // Attach to the device
                var device = Devices[allocation.CardId];
                if (device == null)
                    continue;

                // If the channel name is the same guess we can do it
                if (StringComparer.Ordinal.Equals( channel.DisplayName, allocation.ChannelName ))
                    continue;

                // Now we have to check for the source group - we can record different sources if coming from the same group
                var profile = device.Profile;
                var allowedGroups = profile.FindSource( channel.DisplayName, SourceNameMatchingModes.Name );
                var allocatedGroups = profile.FindSource( allocation.ChannelName, SourceNameMatchingModes.Name );

                // Check for any match - actually this is fairly slow (ok for the given scenario) and not totally correct when using multiple sources with the same name (e.g. Channel 4)
                if (allowedGroups.Any( allowed => allocatedGroups.Any( allocated => allocated.CompareTo( allowed, true ) ) ))
                    continue;

                // Discard
                candidates.Remove( allocation.CardId );

                // Maybe we are done
                if (candidates.Count < 1)
                    break;
            }

            // Report any
            return candidates.FirstOrDefault();
        }

        /// <summary>
        /// Meldet die Aufzeichnungsverzeichnisse in UNC Notation.
        /// </summary>
        /// <returns>Die Liste aller Aufzeichnungsverzeichnisse.</returns>
        public string[] GetRecordingShares()
        {
            // Report
            return new[] { string.Format( @"\\{0}\ArgusTV", Environment.MachineName ) };
        }

        /// <summary>
        /// Meldet alle Verzeichniss zum zeitversetzten Betrachen von Aufzeichnungen in UNC Notation.
        /// </summary>
        /// <returns>Doe Liste der Verzeichnisse.</returns>
        /// <seealso cref="GetRecordingShares"/>
        public string[] GetTimeshiftShares()
        {
            // Just reuse
            return GetRecordingShares();
        }

        /// <summary>
        /// Fordert den Dienst zur Anmeldung auf.
        /// </summary>
        /// <param name="recorderTunerId">Die eindeutige Kennung dieses Dienstes.</param>
        /// <param name="serverHostName">Der Name des Steuerrechners.</param>
        /// <param name="tcpPort">Die Adresse des Steuerdienstes.</param>
        public void Initialize( Guid recorderTunerId, string serverHostName, int tcpPort )
        {
            // Remember
            Identifier = recorderTunerId;

            // Process callback
            using (var agent = new RecorderCallbackServiceAgent( serverHostName, tcpPort ))
                agent.RegisterRecorderTuner( Identifier, Name, GetType().Assembly.GetName().Version.ToString() );
        }

        /// <summary>
        /// Meldet die API Version, mit der dieser Dienst erstellt wurde.
        /// </summary>
        /// <returns>Die verwendete API Version.</returns>
        public int Ping()
        {
            // Report
            return Constants.RecorderApiVersion;
        }

        /// <summary>
        /// Beginnt eine neue Aufzeichnung.
        /// </summary>
        /// <param name="serverHostName">Der Name des ArgusTV Steuerdienstes.</param>
        /// <param name="tcpPort">Die Adresse des Steuerdienstes.</param>
        /// <param name="channelAllocation">Das Gerät, auf dem die Aufzeichnung stattfinden soll.</param>
        /// <param name="startTimeUtc">Der Startzeitpunkt.</param>
        /// <param name="stopTimeUtc">Der Endzeitpunkt.</param>
        /// <param name="recordingProgram">Das aufzuzeichnende Programm.</param>
        /// <param name="suggestedBaseFileName">Ein Vorschlag für den Namen der Aufzeichnungsdatei.</param>
        /// <returns>Gesetzt, wenn der Start eingeleitet wurde.</returns>
        public bool StartRecording( string serverHostName, int tcpPort, CardChannelAllocation channelAllocation, DateTime startTimeUtc, DateTime stopTimeUtc, UpcomingProgram recordingProgram, string suggestedBaseFileName )
        {
            // Enqueue
            return m_activities.GetOrCreate( recordingProgram.UpcomingProgramId, this ).Start( serverHostName, tcpPort, channelAllocation, startTimeUtc, stopTimeUtc, recordingProgram, suggestedBaseFileName );
        }

        /// <summary>
        /// Prüft eine Aufzeichnung und verändert optional den Endzeitpunkt.
        /// </summary>
        /// <param name="channelAllocation">Informationen zur Aufzeichnung.</param>
        /// <param name="recordingProgram">Das betroffene Programm.</param>
        /// <param name="stopTimeUtc">Der neue Endzeitpunkt.</param>
        /// <returns>Gesetzt, wenn die Änderung erfolgreich war.</returns>
        public bool ValidateAndUpdateRecording( CardChannelAllocation channelAllocation, UpcomingProgram recordingProgram, DateTime stopTimeUtc )
        {
            // Locate
            var activity = m_activities.Get( recordingProgram.UpcomingProgramId );
            if (activity == null)
                return false;

            // Forward
            return activity.SetNewStopTime( stopTimeUtc );
        }

        /// <summary>
        /// Beendet den Dienst endgültig.
        /// </summary>
        public void Dispose()
        {
            // Stop timer
            using (m_timer)
                m_timer = null;
        }

        #region Erst einmal nicht unterstützter LIVE Modus

        /// <summary>
        /// Meldet alle <i>LIVE</i> Datenströme.
        /// </summary>
        /// <returns>Die Liste aller Datenströme.</returns>
        public LiveStream[] GetLiveStreams()
        {
            // We do not support LIVE streams
            return new LiveStream[0];
        }

        /// <summary>
        /// Aktiviert den <i>LIVE</i> Modus.
        /// </summary>
        /// <param name="channel">Der gewünschte Sender.</param>
        /// <param name="upcomingRecordingAllocation">Beschreibung der Belegung.</param>
        /// <param name="liveStream">Die Daten zum Datenstrom.</param>
        /// <returns>Das Ergebnis der Operation.</returns>
        public LiveStreamResult TuneLiveStream( Channel channel, CardChannelAllocation upcomingRecordingAllocation, ref LiveStream liveStream )
        {
            // Report
            return LiveStreamResult.NotSupported;
        }

        public bool KeepLiveStreamAlive( LiveStream liveStream )
        {
            throw new NotImplementedException( "KeepLiveStreamAlive" );
        }

        public ServiceTuning GetLiveStreamTuningDetails( LiveStream liveStream )
        {
            throw new NotImplementedException( "GetLiveStreamTuningDetails" );
        }

        public ChannelLiveState[] GetChannelsLiveState( Channel[] channels, LiveStream liveStream )
        {
            throw new NotImplementedException( "GetChannelsLiveState" );
        }

        public void StopLiveStream( LiveStream liveStream )
        {
            throw new NotImplementedException( "StopLiveStream" );
        }

        public bool HasTeletext( LiveStream liveStream )
        {
            throw new NotImplementedException( "HasTeletext" );
        }

        public void StartGrabbingTeletext( LiveStream liveStream )
        {
            throw new NotImplementedException( "StartGrabbingTeletext" );
        }

        public bool IsGrabbingTeletext( LiveStream liveStream )
        {
            throw new NotImplementedException( "IsGrabbingTeletext" );
        }

        public byte[] GetTeletextPageBytes( LiveStream liveStream, int pageNumber, int subPageNumber, out int subPageCount )
        {
            throw new NotImplementedException( "GetTeletextPageBytes" );
        }

        public void StopGrabbingTeletext( LiveStream liveStream )
        {
            throw new NotImplementedException( "StopGrabbingTeletext" );
        }

        #endregion
    }
}
